/*  vsprintf.c                                                  (CRC:f29e47f8)
 *      relative path: ./clib
 *      creation date: 1992
 *      last modified: Wed May 22 21:31:59 1996
 *      header update: Wed Dec 18 07:17:00 1996
 *  
 *       C Library
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  vsprintf.c
 *     virtual printf to a string buffer
 *     author: Frank Barrus
 */

#include <sys/types.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>

/*  the following format characters are supported:
 *	%d	decimal
 *	%u	unsigned decimal
 *	%o	unsigned octal
 *	%x	unsigned hex
 *	%b	unsigned binary
 *	%X	capital unsigned hex
 *	%s	string
 *	%c	character
 *	%n	the number of characters written is stored
 *		in the integer pointed to by the field
 * an optional field width can be specified after the %,
 * and a leading zero implies padding with zeros
 * a negative field width will left justify the field
 */


char *base_binary= "\00201";
char *base_octal = "\01001234567";
char *base_decimal = "\0120123456789";
char *base_hex = "\0200123456789abcdef";
char *base_HEX = "\0200123456789ABCDEF";

/* sprint_int()  takes a string buffer, a number, and a base-string,
 *		 and prints the number to the string, and returns
 *	         the number of digits
 */

static int sprint_int(char *s, long num, char *base)
{
register int q, r;
register int len;
register int x;
	x = 0;
	if (num < 0)  {
		*s++ = '-';
		num = -num;
		x = 1;
	}
	q = num / *base;
	r = num % *base;
	if (q > 0)
		len = sprint_int(s, q, base);
	else
		len = 0;
	s[len] = base[r+1];
	return(len+1+x);
} 



/* sprint_uint() does the same for unsigned integers
 */

static int sprint_uint(char *s, ulong num, char *base)
{
register ulong q, r;
register int len;
	q = num / *base;
	r = num % *base;
	if (q > 0)
		len = sprint_uint(s, q, base);
	else
		len = 0;
	s[len] = base[r+1];
	return(len+1);
} 


/* vsprintf()   takes a string buffer, a format string
 *  		and a pointer to a list of arguments
 *		which are written to the buffer
 *		it returns the number of characters actually written
 */


int vsprintf(char *s, char *fmt, va_list args)
{
register long arg;
register char *sp;
register int len;
register int n;
boolean left;
int size;
char tmp[20];
char padchar;
int i;
	len = 0;
	while(*fmt) if (*fmt == '%') {
		if(fmt[1]=='%') {
			fmt++;
			*s++ = *fmt++;
			continue;
		}	
		if(fmt[1]=='/') {	
			fmt += 2;
			switch(*fmt++) {
			case 'n':
				*s++ = '\n'; break;
			case 'r':
				*s++ = '\r'; break;
			case 'b':
				*s++ = '\b'; break;
			case 't':
				*s++ = '\t'; break;
			case 'a':
				*s++ = '\007'; break;
			case 'f':
				*s++ = '\f'; break;
			}
			continue;
		}
		arg = va_arg(args, long);
		++fmt;
 		n = 0;
		left = FALSE;
		if (*fmt == '-') {
			left = TRUE;
			fmt++;
		}
		if (*fmt == '0') 
			padchar = '0';
		else
			padchar = ' '; 
		while(*fmt >= '0' && *fmt <= '9') 
			n = n * 10 + (*fmt++ - '0');
		size = 0;
		sp = tmp;
		tmp[0] = '\0';
		switch(*fmt++) {
	    	    case 's':
			sp = (char*)arg;
			if(!sp)
				sp = "(null)";
			break;
		    case 'c':
			tmp[0] = (char)arg;
			tmp[1] = '\0';
			break;
		    case 'd':
			size = sprint_int(tmp, arg, base_decimal);
			tmp[size] = '\0';
			break;
		    case 'u':
			size = sprint_uint(tmp, arg, base_decimal);
			tmp[size] = '\0';
			break;
		    case 'b':
			size = sprint_uint(tmp, arg, base_binary);
			tmp[size] = '\0';
			break;
		    case 'o':
			size = sprint_uint(tmp, arg, base_octal);
			tmp[size] = '\0';
			break;
		    case 'x':
			size = sprint_uint(tmp, arg, base_hex);
			tmp[size] = '\0';
			break;
		    case 'X':
			size = sprint_uint(tmp, arg, base_HEX);
			tmp[size] = '\0';
			break;
		    case 'n':
			*(int*)arg = len;
			break;
		}	
		if (!left) {
			i = n - strlen(sp);
			while(i-- > 0) {
				*s++ = padchar;
				len++;
			}
		}
		i = n;
	   	while(*sp && (!n || (i > 0))) {
			*s++ = *sp++;
			len++;
			i--;
		}
		while(left && (i-- > 0)) {
			*s++ = ' ';
			len++;
		}
	} else {
		*s++ = *fmt++;
		len++;
	}
	*s = '\0';
	return(len);
}



/* printf() to a string: */

int sprintf(char *s, char *fmt, ...)
{
va_list args;
int len;
	va_start(args,fmt);
	len = vsprintf(s, fmt, args);
	va_end(args);
	return(len);
}


