/*  dir.c                                                       (CRC:78b524fa)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:04 1996
 *      header update: Wed Dec 18 07:18:05 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* dir.c
 *   simple directory object
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <error.h>
#include <dir.h>


typedef struct dirnode {
	struct dirnode *left,*right;	/* left, right of tree */
	obj_t key;
	obj_t val;
} DirNode;


extern label_t _dir, _mkdir;
label_t _dot, _dotdot;		/* done here, since '.' can't be a label */

ZoneP dir_zone;


static obj_t dir_getnode(DirNode *dir, obj_t name)
{
	if(!dir)
		return(ERR_NOATTR);
	if(name < dir->key)
		return(dir_getnode(dir->left,name));
	else if (name > dir->key)
		return(dir_getnode(dir->right,name));
	else
		return(dir->val);
}

static obj_t dir_setnode(DirNode *dir, obj_t name, ulong val)
{
obj_t oldval;
	if(!dir)
		return(ERR_NOATTR);
	if(name < dir->key)
		return(dir_setnode(dir->left,name,val));
	else if (name > dir->key)
		return(dir_setnode(dir->right,name,val));
	else {
		oldval = dir->val;
		dir->val = val;
		return(oldval);
	}
}


static obj_t dir_addnode(DirNode **dir, obj_t name, ulong val)
{
	if(!(*dir)) {
		*dir = (DirNode*)zalloc(dir_zone);
		(*dir)->key = name;
		(*dir)->val = val;
		(*dir)->left = (*dir)->right = NULL;
		return(OBJ_NULL);
	} 
	if(name < (*dir)->key)
		return(dir_addnode(&(*dir)->left,name,val));
	else if (name > (*dir)->key)
		return(dir_addnode(&(*dir)->right,name,val));
	else 
		return(ERR_EXISTATTR);
}

static obj_t dir_delnode(DirNode **dir, obj_t name)
{
obj_t oldval;
DirNode *right,*left;
	if(!(*dir)) 
		return(ERR_NOATTR);
	if(name < (*dir)->key)
		return(dir_delnode(&(*dir)->left,name));
	else if (name > (*dir)->key)
		return(dir_delnode(&(*dir)->right,name));
	else {
		oldval = (*dir)->val;
		right = (*dir)->right;
		left = (*dir)->left;
		zfree(dir_zone, *dir);
		*dir = left;
		while(*dir)
			dir = &(*dir)->right;
		*dir = right;
		return(oldval);
	}
}


static obj_t dir_getkey(DirNode *dir, int *index)
{
obj_t retval;
	if(!dir)
		return(ERR_EOA);
	retval = dir_getkey(dir->left,index);
	if(retval != ERR_EOA)
		return(retval);
	if(*index==0) 
		return(dir->key);
	(*index)--;
	retval = dir_getkey(dir->right,index);
	return(retval);
}




static obj_t dir_init(ObjFrame *frame, DirNode *dir, obj_t cmd, obj_t name)
{
	frame->data = NULL;
	if(cmd == _mkdir) {
		dir_addnode((DirNode**)&frame->data, _dot,frame->oid);
		dir_addnode((DirNode**)&frame->data, _dotdot,frame->self);
		_(frame->self,_create,name,frame->oid,___);
	}
	return(OFLAG_KERNEL);
}

static obj_t dir_set(ObjFrame *frame, DirNode *dir, obj_t attr, obj_t val)
{
	return(dir_setnode(dir,attr,val));
}

static obj_t dir_mcreate(ObjFrame *frame, DirNode *dir, obj_t attr, obj_t val)
{
	return(dir_addnode((DirNode**)&frame->data,attr,val)); 
}

static obj_t dir_remove(ObjFrame *frame, DirNode *dir, obj_t attr)
{
	return(dir_delnode((DirNode**)&frame->data,attr));
}

static obj_t dir_get(ObjFrame *frame, DirNode *dir, obj_t attr)
{
	return(dir_getnode(dir,attr));
}

static obj_t dir_getattr(ObjFrame *frame, DirNode *dir, int index)
{
	return(dir_getkey(dir,&index));
}

static MethodTable dir_method[] = {
	{&_init, (obj_method)dir_init},
	{&_set, (obj_method)dir_set},
	{&_get, (obj_method)dir_get},
	{&_create, (obj_method)dir_mcreate},
	{&_remove, (obj_method)dir_remove},
	{&_getattr, (obj_method)dir_getattr}};

static ObjClass Class_Dir = {
	dir_method,
	NULLP(AttrTable),
	arraysize(dir_method),
	0};


obj_t dir_parse_path(obj_t cwd, char *path, char *name, boolean create)
{
char *cp;
char buf[256];
obj_t wd;
	if(*path == '/') {
		if(is_kernel()) 
			cwd = kinfo.Root;
		else
			cwd = OBJ_ROOT;
		path++;
	}
	while(1) {
		cp = buf;
		while(*path && *path != '/')
			*cp++ = *path++;
		*cp = '\0';
		if(!*path)
			break;
		wd = err_(cwd,_get,Label(buf),___,___);
		if(err(wd)) {
			if(!create)
				return(wd);
			wd = err_(cwd,_new,_mkdir,Label(buf),___);
			if(err(wd))
				return(wd);
		}
		cwd = wd;
		path++;
	}
	strcpy(name,buf);
	return(cwd);
}


obj_t dir_parse(obj_t cwd, char *path)
{
char buf[256];
	cwd = dir_parse_path(cwd,path,buf,FALSE);
	if(err(cwd) || !buf[0])
		return(cwd);
	cwd = err_(cwd,_get,Label(buf),___,___);
	return(cwd);
}

obj_t dir_create(obj_t cwd, char *path, obj_t val)
{
char buf[120];
obj_t wd;
	wd = dir_parse_path(cwd, path, buf, TRUE);
	if(err(wd))
		return(wd);
	wd = err_(wd, _create, Label(buf), val, ___);
	if(err(wd))
		return(wd);
	else
		return(val);
}
 
	
void Dir_init(void)
{
	dir_zone = (ZoneP)zinit(sizeof(DirNode),"DirNode");
	_dot = Label(".");
	_dotdot = Label("..");
	class_dir = _(class_object, _class, Ptr(&Class_Dir), ___, ___);
	kinfo.Root = _(class_dir, _new, ___, ___, ___);
	_(kinfo.Root,_create,_dot,kinfo.Root,___);
	_(kinfo.Root,_create,_dotdot,kinfo.Root,___);
	dir_create(kinfo.Root, "/class/Dir", class_dir);
	dir_create(kinfo.Root, "/class/Object", class_object);
	dir_create(kinfo.Root, "/class/Class", class_class);
}

